"""
This file contains a number of helper utilities to set up our various
experiments with less code.
"""
import numpy as np
from pydrake.all import (BallRpyJoint, Box, CoulombFriction, Cylinder,
                         PrismaticJoint, RevoluteJoint, RigidTransform,
                         SpatialInertia, Sphere, UnitInertia)


def AddShape(plant, shape, name, mass=1, mu=1, color=[.5, .5, .9, 1.0]):
    instance = plant.AddModelInstance(name)
    # TODO: Add a method to UnitInertia that accepts a geometry shape (unless
    # that dependency is somehow gross) and does this.
    if isinstance(shape, Box):
        inertia = UnitInertia.SolidBox(shape.width(), shape.depth(),
                                       shape.height())
    elif isinstance(shape, Cylinder):
        inertia = UnitInertia.SolidCylinder(shape.radius(), shape.length())
    elif isinstance(shape, Sphere):
        inertia = UnitInertia.SolidSphere(shape.radius())
    else:
        raise RunTimeError(
            f"need to write the unit inertia for shapes of type {shape}")
    body = plant.AddRigidBody(
        name, instance,
        SpatialInertia(mass=mass,
                       p_PScm_E=np.array([0., 0., 0.]),
                       G_SP_E=inertia))
    if plant.geometry_source_is_registered():
        if isinstance(shape, Box):
            plant.RegisterCollisionGeometry(
                body, RigidTransform(),
                Box(shape.width() - 0.001,
                    shape.depth() - 0.001,
                    shape.height() - 0.001), name, CoulombFriction(mu, mu))
            i = 0
            for x in [-shape.width() / 2.0, shape.width() / 2.0]:
                for y in [-shape.depth() / 2.0, shape.depth() / 2.0]:
                    for z in [-shape.height() / 2.0, shape.height() / 2.0]:
                        plant.RegisterCollisionGeometry(
                            body, RigidTransform([x, y, z]),
                            Sphere(radius=1e-7), f"contact_sphere{i}",
                            CoulombFriction(mu, mu))
                        i += 1
        else:
            plant.RegisterCollisionGeometry(body, RigidTransform(), shape, name,
                                            CoulombFriction(mu, mu))

        plant.RegisterVisualGeometry(body, RigidTransform(), shape, name, color)

    return instance


# https://github.com/RobotLocomotion/drake/issues/14949
def AddFloatingRpyJoint(plant, frame, instance, use_ball_rpy=True):
    inertia = SpatialInertia(mass=0,
                             p_PScm_E=[0., 0., 0.],
                             G_SP_E=UnitInertia(0, 0, 0))
    x_body = plant.AddRigidBody("x", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("x", plant.world_frame(), x_body.body_frame(),
                       [1, 0, 0]))
    y_body = plant.AddRigidBody("y", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("y", x_body.body_frame(), y_body.body_frame(),
                       [0, 1, 0]))
    z_body = plant.AddRigidBody("z", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("z", y_body.body_frame(), z_body.body_frame(),
                       [0, 0, 1]))
    if use_ball_rpy:
        plant.AddJoint(BallRpyJoint("ball", z_body.body_frame(), frame))
    else:
        # RollPitchYaw is body z-y-x
        rz_body = plant.AddRigidBody("rz", instance, inertia)
        plant.AddJoint(
            RevoluteJoint("rz", z_body.body_frame(), rz_body.body_frame(),
                          [0, 0, 1]))
        ry_body = plant.AddRigidBody("ry", instance, inertia)
        plant.AddJoint(
            RevoluteJoint("ry", rz_body.body_frame(), ry_body.body_frame(),
                          [0, 1, 0]))
        plant.AddJoint(
            RevoluteJoint("rx", ry_body.body_frame(), frame, [1, 0, 0]))



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestIlqrDriving(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test1_discrete_dynamics(self):
        """Test discrete_dynamics"""
        # load locals
        f = self.notebook_locals['discrete_dynamics']

        f_target = [
            np.array(
                [0.14181286, 0.81062987, 0.54580749, 0.77731476, 1.02810156]),
            np.array([0.131663, 0.300996, 0.57034388, 0.71732411, 0.81033267]),
            np.array(
                [0.33233256, 0.91916814, 0.27416228, 0.45461388, 0.99126091]),
            np.array(
                [1.02770659, 0.27770476, 0.56066881, 0.96146963, 0.20821043]),
            np.array([0.71706372, 0.47773599, 0.22401885, 0.538506, 0.40897373])
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            x = np.random.rand(5)
            u = np.random.rand(2)
            f_eval = f(x, u)
            np.testing.assert_array_almost_equal(
                f_eval,
                f_target[i],
                err_msg='The discrete dynamics are incorrect. '
                f'Expected output {f_target[i]} with x={x} and u={u}')

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test2_rollout(self):
        """Test rollout"""
        # load locals
        f = self.notebook_locals['rollout']

        f_target_test_target = [
            0.36531949939751773,
            2.25991775105549,
            -0.05889795147375571,
            0.5476245815855634,
            0.8066768897676444,
        ]

        np.random.seed(7)
        n_rands = 5
        for i in range(n_rands):
            N = 5
            n_u = 2
            x0 = np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            f_eval_test = np.random.randn(N) @ f(x0, u_trj) @ np.random.randn(N)
            msg = f'The discrete dynamics are incorrect.' + \
                'Expected output {f_target_test_target[i]} got {f_eval_test}'
            self.assertAlmostEqual(f_eval_test,
                                   f_target_test_target[i],
                                   msg=msg)

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test3_cost_trj(self):
        """Test cost_trj"""
        # load locals
        f = self.notebook_locals['cost_trj']

        f_target = [
            29.946731429035616, 34.805336401404986, 27.626872505986434,
            33.25730851334137, 33.83897480138296
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            N = 5
            n_u = 2
            n_x = 5
            x0 = np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            x_trj = np.random.randn(N, n_x) * 0.4
            f_eval = f(x_trj, u_trj)
            np.testing.assert_almost_equal(
                f_eval,
                f_target[i],
                err_msg='The cost_trj function is incorrect. '
                f'Expected output {f_target[i]} with x_trj=\n{x_trj}'
                f' and u=\n{u_trj}')

    @weight(5)
    @timeout_decorator.timeout(1.)
    def test4_Q_terms(self):
        """Test Q_terms"""
        # load locals
        Q_terms = self.notebook_locals['Q_terms']
        derivs = self.notebook_locals['derivs']

        np.random.seed(7)
        x = np.random.rand(5)
        u = np.random.rand(2)

        l_x, l_u, l_xx, l_ux, l_uu, f_x, f_u = derivs.stage(x, u)
        V_x, V_xx = derivs.final(x + 0.1)
        Q_x, Q_u, Q_xx, Q_ux, Q_uu = Q_terms(l_x, l_u, l_xx, l_ux, l_uu, f_x,
                                             f_u, V_x, V_xx)

        Q_x_test_val = (Q_x @ np.random.randn(Q_x.shape[0]))
        Q_u_test_val = (Q_u @ np.random.randn(Q_u.shape[0]))
        Q_xx_test_val = (np.random.randn(Q_xx.shape[0]) @ Q_xx
                         @ np.random.randn(Q_xx.shape[0]))
        Q_ux_test_val = (np.random.randn(Q_ux.shape[0]) @ Q_ux
                         @ np.random.randn(Q_ux.shape[1]))
        Q_uu_test_val = (np.random.randn(Q_uu.shape[0]) @ Q_uu
                         @ np.random.randn(Q_uu.shape[0]))

        Q_x_test_target = 1.1817013977437874
        Q_u_test_target = -0.1716142906672626
        Q_xx_test_target = 0.5337352394014466
        Q_ux_test_target = 0.5156605288974515
        Q_uu_test_target = -0.28166353613679074

        self.assertAlmostEqual(Q_x_test_val,
                               Q_x_test_target,
                               msg="Incorrect Q_x")
        self.assertAlmostEqual(Q_u_test_val,
                               Q_u_test_target,
                               msg="Incorrect Q_u")
        self.assertAlmostEqual(Q_xx_test_val,
                               Q_xx_test_target,
                               msg="Incorrect Q_xx")
        self.assertAlmostEqual(Q_ux_test_val,
                               Q_ux_test_target,
                               msg="Incorrect Q_ux")
        self.assertAlmostEqual(Q_uu_test_val,
                               Q_uu_test_target,
                               msg="Incorrect Q_uu")

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test5_gains(self):
        """Test gains"""
        # load locals
        gains = self.notebook_locals['gains']

        np.random.seed(8)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        k, K = gains(Q_uu, Q_u, Q_ux)

        test_inputs = 10 * np.random.randn(k.shape[0], 5)

        k_test_val = (k @ test_inputs).sum()
        K_test_val = (test_inputs.T @ K).sum()

        self.assertAlmostEqual(k_test_val,
                               10.712481345027399,
                               msg="Incorrect k gain")

        self.assertAlmostEqual(K_test_val,
                               -12.949530490536842,
                               msg="Incorrect K gain")

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test6_V_terms(self):
        """Test V_terms"""
        # load locals
        V_terms = self.notebook_locals['V_terms']
        gains = self.notebook_locals['gains']

        np.random.seed(7)
        Q_x = np.random.randn(5)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        Q_xx = np.random.randn(5, 5)
        Q_xx = 0.5 * (Q_xx + Q_xx.T)
        k, K = gains(Q_uu, Q_u, Q_ux)

        V_x, V_xx = V_terms(Q_x, Q_u, Q_xx, Q_ux, Q_uu, K, k)
        test_inputs = np.random.randn(V_x.shape[0])

        V_x_test_val = V_x @ test_inputs
        V_xx_test_val = test_inputs @ V_xx @ test_inputs

        self.assertAlmostEqual(V_x_test_val,
                               -2.113959925580339,
                               msg="Incorrect V_x")

        self.assertAlmostEqual(V_xx_test_val,
                               35.931091753768,
                               msg="Incorrect V_xx")

        # this tests whether the student simplified the V_terms expression
        k2, K2 = 10 * np.random.randn(*k.shape), 10 * np.random.randn(*K.shape)
        V_x2, V_xx2 = V_terms(Q_x, Q_u, Q_xx, Q_ux, Q_uu, K2, k2)

        V_x2_test_val = V_x2 @ test_inputs
        V_xx2_test_val = test_inputs @ V_xx2 @ test_inputs

        self.assertNotAlmostEqual(
            V_x2_test_val,
            2.484304411129088,
            msg="""It appears you may have simplified the expression for V_x.
            Do not do this""")
        self.assertNotAlmostEqual(
            V_xx2_test_val,
            265.0164583112695,
            msg="""It appears you may have simplified the expression for V_xx.
            Do not do this""")

        self.assertAlmostEqual(V_x2_test_val,
                               -6.043415622930539,
                               msg="Incorrect V_x")

        self.assertAlmostEqual(V_xx2_test_val,
                               -52.01732694080634,
                               msg="Incorrect V_xx")

    @weight(2)
    @timeout_decorator.timeout(1.)
    def test7_forward_pass(self):
        """Test forward_pass"""
        # load locals
        forward_pass = self.notebook_locals['forward_pass']

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)
        k_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1])
        K_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1], x_trj.shape[1])
        x_trj_new, u_trj_new = forward_pass(x_trj, u_trj, k_trj, K_trj)

        x_test = np.random.randn(x_trj_new.shape[0])
        u_test = np.random.randn(u_trj_new.shape[0])

        x_trj_new_test_val = x_test @ x_trj_new @ x_test
        u_trj_new_test_val = (u_test @ u_trj_new).sum()

        self.assertAlmostEqual(x_trj_new_test_val,
                               -1.0615491271322297,
                               msg="Incorrect x_trj_new")

        self.assertAlmostEqual(u_trj_new_test_val,
                               -10.649763284585609,
                               msg="Incorrect u_trj_new")

    @weight(2)
    @timeout_decorator.timeout(1.)
    def test8_backward_pass(self):
        """Test backward_pass"""
        # load locals
        backward_pass = self.notebook_locals['backward_pass']

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        k = np.random.randn(2)
        K = np.random.randn(2, 5)
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)

        k_trj, K_trj, _ = backward_pass(x_trj, u_trj, 100.0)

        r_test = 10 * np.random.randn(k_trj.shape[1])
        l_test = 10 * np.random.randn(k_trj.shape[0])

        k_trj_test_val = l_test @ k_trj @ r_test
        K_trj_test_val = np.einsum("i, ijk, j -> k", l_test, K_trj,
                                   r_test).sum()

        self.assertAlmostEqual(k_trj_test_val,
                               -5.380572857461254,
                               msg="Incorrect k_trj")

        self.assertAlmostEqual(K_trj_test_val,
                               3.770894929116537,
                               msg="Incorrect K_trj")



import os

from pydrake.all import (JointActuatorIndex, JointIndex, namedview)

running_as_test = False


def set_running_as_test(value):
    global running_as_test
    running_as_test = value


def FindResource(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def Rgba2Hex(rgb):
    """
    Turn a list of R,G,B elements (any indexable list of >= 3 elements will
    work), where each element is specified on range [0., 1.], into the
    equivalent 24-bit value 0xRRGGBB.
    """
    val = 0
    for i in range(3):
        val += (256**(2 - i)) * int(255 * rgb[i])
    return val


# TODO(russt): promote these to drake (and make a version with model_instance)


def MakeNamedViewPositions(mbp, view_name, add_suffix_if_single_position=False):
    names = [None] * mbp.num_positions()
    for ind in range(mbp.num_joints()):
        joint = mbp.get_joint(JointIndex(ind))
        if joint.num_positions() == 1 and not add_suffix_if_single_position:
            names[joint.position_start()] = joint.name()
        else:
            for i in range(joint.num_positions()):
                names[joint.position_start() + i] = \
                    f"{joint.name()}_{joint.position_suffix(i)}"
    for ind in mbp.GetFloatingBaseBodies():
        body = mbp.get_body(ind)
        start = body.floating_positions_start()
        for i in range(7 if body.has_quaternion_dofs() else 6):
            names[start
                  + i] = f"{body.name()}_{body.floating_position_suffix(i)}"
    return namedview(view_name, names)


def MakeNamedViewVelocities(mbp,
                            view_name,
                            add_suffix_if_single_velocity=False):
    names = [None] * mbp.num_velocities()
    for ind in range(mbp.num_joints()):
        joint = mbp.get_joint(JointIndex(ind))
        if joint.num_velocities() == 1 and not add_suffix_if_single_velocity:
            names[joint.velocity_start()] = joint.name()
        else:
            for i in range(joint.num_velocities()):
                names[joint.velocity_start() + i] = \
                    f"{joint.name()}_{joint.velocity_suffix(i)}"
    for ind in mbp.GetFloatingBaseBodies():
        body = mbp.get_body(ind)
        start = body.floating_velocities_start() - mbp.num_positions()
        for i in range(6):
            names[start
                  + i] = f"{body.name()}_{body.floating_velocity_suffix(i)}"
    return namedview(view_name, names)


def MakeNamedViewState(mbp, view_name):
    # TODO: this could become a nested named view, pending
    # https://github.com/RobotLocomotion/drake/pull/14973
    pview = MakeNamedViewPositions(mbp, f"{view_name}_pos", True)
    vview = MakeNamedViewVelocities(mbp, f"{view_name}_vel", True)
    return namedview(view_name, pview.get_fields() + vview.get_fields())


def MakeNamedViewActuation(mbp, view_name):
    names = [None] * mbp.get_actuation_input_port().size()
    for ind in range(mbp.num_actuators()):
        actuator = mbp.get_joint_actuator(JointActuatorIndex(ind))
        assert actuator.num_inputs() == 1
        names[actuator.input_start()] = actuator.name()
    return namedview(view_name, names)



from functools import partial
import time

import numpy as np

from pydrake.geometry import Cylinder, Rgba, Sphere
from pydrake.perception import PointCloud, Fields, BaseField
from pydrake.solvers import BoundingBoxConstraint

# imports for interact
from inspect import signature, Parameter
from ipywidgets.widgets.interaction import (_get_min_max_value,
                                            _yield_abbreviations_for_parameter)

# imports for the pose sliders
from collections import namedtuple
from pydrake.common.value import AbstractValue
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.systems.framework import LeafSystem, PublishEvent

# imports for the joint sliders
from pydrake.multibody.tree import JointIndex
from pydrake.systems.framework import PublishEvent

from underactuated import running_as_notebook

# Some GUI code that will be moved into Drake.


def interact(meshcat, callback, **kwargs):
    """
    A poor-person's implementation of a functionality like ipywidgets interact
    https://ipywidgets.readthedocs.io/en/latest/examples/Using%20Interact.html#Basic-interact
    """
    values = {}

    # The following code was adapted from https://github.com/jupyter-widgets/ipywidgets/blob/9326d28c474cb6d2b58876d469489a73e98f903a/python/ipywidgets/ipywidgets/widgets/interaction.py  # noqa
    def find_abbreviations(callback, kwargs):
        """
        Find the abbreviations for the given function and kwargs.
        Return (name, abbrev, default) tuples.
        """
        new_kwargs = []
        try:
            sig = signature(callback)
        except (ValueError, TypeError):
            # can't inspect, no info from function; only use kwargs
            return [(key, value, value) for key, value in kwargs.items()]

        for param in sig.parameters.values():
            for name, value, default in _yield_abbreviations_for_parameter(
                    param, kwargs):
                if value is Parameter.empty:
                    raise ValueError(
                        'cannot find widget or abbreviation for argument: {!r}'.
                        format(name))
                new_kwargs.append((name, value, default))
        return new_kwargs

    new_kwargs = find_abbreviations(callback, kwargs)

    for name, abbrev, default in new_kwargs:
        kw = {}
        kw['value'] = None if default is Parameter.empty else default
        if isinstance(abbrev, tuple):
            kw['step'] = abbrev[2] if len(abbrev) != 3 else None
            kw["min"], kw["max"], kw["value"] = _get_min_max_value(
                abbrev[0], abbrev[1], **kw)
            if kw['step'] is None:
                kw['step'] = 0.1
            meshcat.AddSlider(name, **kw)
        else:
            raise ValueError("This case is not implemented yet")
            # It might be simple.  I just haven't tried!
        values[name] = kw['value']

    def update_values():
        changed = False
        for name in values:
            v = meshcat.GetSliderValue(name)
            changed |= v != values[name]
            values[name] = v
        return changed

    # Always call it once.
    callback(**values)

    if not running_as_notebook:
        return

    print("Press the 'Stop Interacting' button in Meshcat to continue.")
    meshcat.AddButton("Stop Interacting")
    while meshcat.GetButtonClicks("Stop Interacting") < 1:
        if update_values():
            callback(**values)
        time.sleep(.1)

    meshcat.DeleteButton("Stop Interacting")
    for name in values:
        meshcat.DeleteSlider(name)


class MeshcatSliders(LeafSystem):
    """
    A system that outputs the ``value``s from meshcat sliders.

    .. pydrake_system::

      name: MeshcatSliderSystem
      output_ports:
      - slider_group_0
      - ...
      - slider_group_{N-1}
    """

    def __init__(self, meshcat, slider_names):
        """
        An output port is created for each element in the list `slider_names`.
        Each element of `slider_names` must itself be an iterable collection
        (list, tuple, set, ...) of strings, with the names of sliders that have
        *already* been added to Meshcat via Meshcat.AddSlider().

        The same slider may be used in multiple ports.
        """
        LeafSystem.__init__(self)

        self._meshcat = meshcat
        self._sliders = slider_names
        for i, slider_iterable in enumerate(self._sliders):
            port = self.DeclareVectorOutputPort(
                f"slider_group_{i}", len(slider_iterable),
                partial(self.DoCalcOutput, port_index=i))
            port.disable_caching_by_default()

    def DoCalcOutput(self, context, output, port_index):
        for i, slider in enumerate(self._sliders[port_index]):
            output[i] = self._meshcat.GetSliderValue(slider)


class MeshcatPoseSliders(LeafSystem):
    """
    Provides a set of ipywidget sliders (to be used in a Jupyter notebook) with
    one slider for each of roll, pitch, yaw, x, y, and z.  This can be used,
    for instance, as an interface to teleoperate the end-effector of a robot.

    .. pydrake_system::

        name: PoseSliders
        output_ports:
        - pose
    """
    # TODO(russt): Use namedtuple defaults parameter once we are Python >= 3.7.
    Visible = namedtuple("Visible", ("roll", "pitch", "yaw", "x", "y", "z"))
    Visible.__new__.__defaults__ = (True, True, True, True, True, True)
    MinRange = namedtuple("MinRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MinRange.__new__.__defaults__ = (-np.pi, -np.pi, -np.pi, -1.0, -1.0, -1.0)
    MaxRange = namedtuple("MaxRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MaxRange.__new__.__defaults__ = (np.pi, np.pi, np.pi, 1.0, 1.0, 1.0)
    Value = namedtuple("Value", ("roll", "pitch", "yaw", "x", "y", "z"))
    Value.__new__.__defaults__ = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

    def __init__(self,
                 meshcat,
                 visible=Visible(),
                 min_range=MinRange(),
                 max_range=MaxRange(),
                 value=Value()):
        """
        Args:
            meshcat: A Meshcat instance.
            visible: An object with boolean elements for 'roll', 'pitch',
                     'yaw', 'x', 'y', 'z'; the intention is for this to be the
                     PoseSliders.Visible() namedtuple.  Defaults to all true.
            min_range, max_range, value: Objects with float values for 'roll',
                      'pitch', 'yaw', 'x', 'y', 'z'; the intention is for the
                      caller to use the PoseSliders.MinRange, MaxRange, and
                      Value namedtuples.  See those tuples for default values.
        """
        LeafSystem.__init__(self)
        port = self.DeclareAbstractOutputPort(
            "pose", lambda: AbstractValue.Make(RigidTransform()),
            self.DoCalcOutput)

        # The widgets themselves have undeclared state.  For now, we accept it,
        # and simply disable caching on the output port.
        # TODO(russt): consider implementing the more elaborate methods seen
        # in, e.g., LcmMessageSubscriber.
        port.disable_caching_by_default()

        self._meshcat = meshcat
        self._visible = visible
        self._value = list(value)

        for i in range(6):
            if visible[i]:
                meshcat.AddSlider(min=min_range[i],
                                  max=max_range[i],
                                  value=value[i],
                                  step=0.01,
                                  name=value._fields[i])

    def __del__(self):
        for s in ['roll', 'pitch', 'yaw', 'x', 'y', 'z']:
            if visible[s]:
                self._meshcat.DeleteSlider(s)

    def SetPose(self, pose):
        """
        Sets the current value of the sliders.

        Args:
            pose: Any viable argument for the RigidTransform
                  constructor.
        """
        tf = RigidTransform(pose)
        self.SetRpy(RollPitchYaw(tf.rotation()))
        self.SetXyz(tf.translation())

    def SetRpy(self, rpy):
        """
        Sets the current value of the sliders for roll, pitch, and yaw.

        Args:
            rpy: An instance of drake.math.RollPitchYaw
        """
        self._value[0] = rpy.roll_angle()
        self._value[1] = rpy.pitch_angle()
        self._value[2] = rpy.yaw_angle()
        for i in range(3):
            if self._visible[i]:
                self._meshcat.SetSliderValue(self._visible._fields[i],
                                             self._value[i])

    def SetXyz(self, xyz):
        """
        Sets the current value of the sliders for x, y, and z.

        Args:
            xyz: A 3 element iterable object with x, y, z.
        """
        self._value[3:] = xyz
        for i in range(3, 6):
            if self._visible[i]:
                self._meshcat.SetSliderValue(self._visible._fields[i],
                                             self._value[i])

    def _update_values(self):
        changed = False
        for i in range(6):
            if self._visible[i]:
                old_value = self._value[i]
                self._value[i] = self._meshcat.GetSliderValue(
                    self._visible._fields[i])
                changed = changed or self._value[i] != old_value
        return changed

    def _get_transform(self):
        return RigidTransform(
            RollPitchYaw(self._value[0], self._value[1], self._value[2]),
            self._value[3:])

    def DoCalcOutput(self, context, output):
        """Constructs the output values from the sliders."""
        self._update_values()
        output.set_value(self._get_transform())

    def Run(self, publishing_system, root_context, callback):
        # Calls callback(root_context, pose), then publishing_system.Publish()
        # each time the sliders change value.
        if not running_as_notebook:
            return

        publishing_context = publishing_system.GetMyContextFromRoot(
            root_context)

        print("Press the 'Stop PoseSliders' button in Meshcat to continue.")
        self._meshcat.AddButton("Stop PoseSliders")
        while self._meshcat.GetButtonClicks("Stop PoseSliders") < 1:
            if self._update_values():
                callback(root_context, self._get_transform())
                publishing_system.Publish(publishing_context)
            time.sleep(.1)

        self._meshcat.DeleteButton("Stop PoseSliders")


class WsgButton(LeafSystem):

    def __init__(self, meshcat):
        LeafSystem.__init__(self)
        port = self.DeclareVectorOutputPort("wsg_position", 1,
                                            self.DoCalcOutput)
        port.disable_caching_by_default()
        self._meshcat = meshcat
        self._button = "Open/Close Gripper"
        meshcat.AddButton(self._button)

    def __del__(self):
        self._meshcat.DeleteButton(self._button)

    def DoCalcOutput(self, context, output):
        position = 0.107  # open
        if (self._meshcat.GetButtonClicks(self._button) % 2) == 1:
            position = 0.002  # close
        output.SetAtIndex(0, position)


# TODO(russt): Add floating base support (as in manipulation.jupyter_widgets
# MakeJointSlidersThatPublish)
class MeshcatJointSliders(LeafSystem):
    """
    Adds one slider per joint of the MultibodyPlant.  Any positions that are
    not associated with joints (e.g. floating-base "mobilizers") are held
    constant at the default value obtained from robot.CreateDefaultContext().

    .. pydrake_system::

        name: JointSliders
        output_ports:
        - positions

    In addition to being used inside a Diagram that is being simulated with
    Simulator, this class also offers a `Run` method that runs its own simple
    event loop, querying the slider values and calling `Publish`.  It does not
    simulate any state dynamics.
    """

    def __init__(self,
                 meshcat,
                 plant,
                 root_context=None,
                 lower_limit=-10.,
                 upper_limit=10.,
                 resolution=0.01):
        """
        Creates an meshcat slider for each joint in the plant.

        Args:
            meshcat:      A Meshcat instance.
            plant:        A MultibodyPlant. publishing_system: The System whose
                          Publish method will be called.  Can be the entire
                          Diagram, but can also be a subsystem.
            root_context: A mutable root Context of the Diagram containing the
                          ``plant``; we will extract the subcontext's using
                          `GetMyContextFromRoot`.
            lower_limit:  A scalar or vector of length robot.num_positions().
                          The lower limit of the slider will be the maximum
                          value of this number and any limit specified in the
                          Joint.

            upper_limit:  A scalar or vector of length robot.num_positions().
                          The upper limit of the slider will be the minimum
                          value of this number and any limit specified in the
                          Joint.

            resolution:   A scalar or vector of length robot.num_positions()
                          that specifies the step argument of the FloatSlider.
        """
        LeafSystem.__init__(self)

        def _broadcast(x, num):
            x = np.array(x)
            assert len(x.shape) <= 1
            return np.array(x) * np.ones(num)

        lower_limit = _broadcast(lower_limit, plant.num_positions())
        upper_limit = _broadcast(upper_limit, plant.num_positions())
        resolution = _broadcast(resolution, plant.num_positions())

        self._meshcat = meshcat
        self._plant = plant
        plant_context = plant.GetMyContextFromRoot(root_context) if \
            root_context else plant.CreateDefaultContext()

        self._sliders = {}
        self._positions = plant.GetPositions(plant_context)
        slider_num = 0
        for i in range(plant.num_joints()):
            joint = plant.get_joint(JointIndex(i))
            low = joint.position_lower_limits()
            upp = joint.position_upper_limits()
            for j in range(joint.num_positions()):
                index = joint.position_start() + j
                description = joint.name()
                if joint.num_positions() > 1:
                    description += '_' + joint.position_suffix(j)
                meshcat.AddSlider(value=self._positions[index],
                                  min=max(low[j], lower_limit[slider_num]),
                                  max=min(upp[j], upper_limit[slider_num]),
                                  step=resolution[slider_num],
                                  name=description)
                self._sliders[index] = description
                slider_num += 1

        port = self.DeclareVectorOutputPort("positions", plant.num_positions(),
                                            self.DoCalcOutput)
        port.disable_caching_by_default()

    def DoCalcOutput(self, context, output):
        output.SetFromVector(self._positions)
        for i, s in self._sliders.items():
            output[i] = self._meshcat.GetSliderValue(s)

    def Run(self, publishing_system, root_context, callback=None):
        """
        Args:
            publishing_system:  The system to call publish on.  Probably a
                          MeshcatVisualizer.
            root_context: A mutable root Context of the Diagram containing both
                          the ``plant`` and the ``publishing_system``; we will
                          extract the subcontext's using `GetMyContextFromRoot`.
            callback: callback(plant_context) will be called whenever the
                      slider values change.
        """
        if not running_as_notebook:
            return
        print("Press the 'Stop JointSliders' button in Meshcat to continue.")
        self._meshcat.AddButton("Stop JointSliders")

        plant_context = self._plant.GetMyContextFromRoot(root_context)
        publishing_context = publishing_system.GetMyContextFromRoot(
            root_context)

        publishing_system.Publish(publishing_context)
        while self._meshcat.GetButtonClicks("Stop JointSliders") < 1:
            old_positions = self._plant.GetPositions(plant_context)
            positions = self._positions
            for i, s in self._sliders.items():
                positions[i] = self._meshcat.GetSliderValue(s)
            if not np.array_equal(positions, old_positions):
                self._plant.SetPositions(plant_context, positions)
                if callback:
                    callback(plant_context)
                publishing_system.Publish(publishing_context)
            time.sleep(.1)

        self._meshcat.DeleteButton("Stop JointSliders")


def AddMeshcatTriad(meshcat,
                    path,
                    length=.25,
                    radius=0.01,
                    opacity=1.,
                    X_PT=RigidTransform()):
    meshcat.SetTransform(path, X_PT)
    # x-axis
    X_TG = RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2),
                          [length / 2., 0, 0])
    meshcat.SetTransform(path + "/x-axis", X_TG)
    meshcat.SetObject(path + "/x-axis", Cylinder(radius, length),
                      Rgba(1, 0, 0, opacity))

    # y-axis
    X_TG = RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2),
                          [0, length / 2., 0])
    meshcat.SetTransform(path + "/y-axis", X_TG)
    meshcat.SetObject(path + "/y-axis", Cylinder(radius, length),
                      Rgba(0, 1, 0, opacity))

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.])
    meshcat.SetTransform(path + "/z-axis", X_TG)
    meshcat.SetObject(path + "/z-axis", Cylinder(radius, length),
                      Rgba(0, 0, 1, opacity))


def draw_open3d_point_cloud(meshcat,
                            path,
                            pcd,
                            normals_scale=0.0,
                            point_size=0.001):
    pts = np.asarray(pcd.points)
    assert (pcd.has_colors())  # TODO(russt): handle this case better
    cloud = PointCloud(pts.shape[0], Fields(BaseField.kXYZs | BaseField.kRGBs))
    cloud.mutable_xyzs()[:] = pts.T
    cloud.mutable_rgbs()[:] = 255 * np.asarray(pcd.colors).T
    meshcat.SetObject(path, cloud, point_size=point_size)
    if pcd.has_normals() and normals_scale > 0.0:
        assert ('need to implement LineSegments in meshcat c++')
        normals = np.asarray(pcd.normals)
        vertices = np.hstack(
            (pts, pts + normals_scale * normals)).reshape(-1, 3).T
        meshcat["normals"].set_object(
            g.LineSegments(g.PointsGeometry(vertices),
                           g.MeshBasicMaterial(color=0x000000)))


def plot_surface(meshcat,
                 path,
                 X,
                 Y,
                 Z,
                 rgba=Rgba(.87, .6, .6, 1.0),
                 wireframe=False,
                 wireframe_line_width=1.0):
    (rows, cols) = Z.shape
    assert (np.array_equal(X.shape, Y.shape))
    assert (np.array_equal(X.shape, Z.shape))

    vertices = np.empty((rows * cols, 3), dtype=np.float32)
    vertices[:, 0] = X.reshape((-1))
    vertices[:, 1] = Y.reshape((-1))
    vertices[:, 2] = Z.reshape((-1))

    # Vectorized faces code from https://stackoverflow.com/questions/44934631/making-grid-triangular-mesh-quickly-with-numpy  # noqa
    faces = np.empty((rows - 1, cols - 1, 2, 3), dtype=np.uint32)
    r = np.arange(rows * cols).reshape(rows, cols)
    faces[:, :, 0, 0] = r[:-1, :-1]
    faces[:, :, 1, 0] = r[:-1, 1:]
    faces[:, :, 0, 1] = r[:-1, 1:]
    faces[:, :, 1, 1] = r[1:, 1:]
    faces[:, :, :, 2] = r[1:, :-1, None]
    faces.shape = (-1, 3)

    # TODO(Russ): support per vertex / Colormap colors.
    meshcat.SetTriangleMesh(path, vertices.T, faces.T, rgba, wireframe,
                            wireframe_line_width)


def plot_mathematical_program(meshcat,
                              path,
                              prog,
                              X,
                              Y,
                              result=None,
                              point_size=0.05):
    assert prog.num_vars() == 2
    assert X.size == Y.size

    N = X.size
    values = np.vstack((X.reshape(-1), Y.reshape(-1)))
    costs = prog.GetAllCosts()

    # Vectorized multiply for the quadratic form.
    # Z = (D*np.matmul(Q,D)).sum(0).reshape(nx, ny)

    if costs:
        Z = prog.EvalBindingVectorized(costs[0], values)
        for b in costs[1:]:
            Z = Z + prog.EvalBindingVectorized(b, values)

    cv = f"{path}/constraints"
    for binding in prog.GetAllConstraints():
        if isinstance(binding.evaluator(), BoundingBoxConstraint):
            c = binding.evaluator()
            var_indices = [
                int(prog.decision_variable_index()[v.get_id()])
                for v in binding.variables()
            ]
            satisfied = np.array(
                c.CheckSatisfiedVectorized(values[var_indices, :],
                                           0.001)).reshape(1, -1)
            if costs:
                Z[~satisfied] = np.nan

            v = f"{cv}/{type(c).__name__}"
            Zc = np.zeros(Z.shape)
            Zc[satisfied] = np.nan
            plot_surface(meshcat,
                         v,
                         X,
                         Y,
                         Zc.reshape(X.shape),
                         rgba=Rgba(1.0, .2, .2, 1.0),
                         wireframe=True)
        else:
            Zc = prog.EvalBindingVectorized(binding, values)
            evaluator = binding.evaluator()
            low = evaluator.lower_bound()
            up = evaluator.upper_bound()
            cvb = f"{cv}/{type(evaluator).__name__}"
            for index in range(Zc.shape[0]):
                # TODO(russt): Plot infeasible points in a different color.
                infeasible = np.logical_or(Zc[index, :] < low[index],
                                           Zc[index, :] > up[index])
                plot_surface(meshcat,
                             f"{cvb}/{index}",
                             X,
                             Y,
                             Zc[index, :].reshape(X.shape),
                             rgba=Rgba(1.0, .3, 1.0, 1.0),
                             wireframe=True)

    if costs:
        plot_surface(meshcat,
                     f"{path}/objective",
                     X,
                     Y,
                     Z.reshape(X.shape),
                     rgba=Rgba(.3, 1.0, .3, 1.0),
                     wireframe=True)

    if result:
        v = f"{path}/solution"
        meshcat.SetObject(v, Sphere(point_size), Rgba(.3, 1.0, .3, 1.0))
        x_solution = result.get_x_val()
        meshcat.SetTransform(
            v,
            RigidTransform(
                [x_solution[0], x_solution[1],
                 result.get_optimal_cost()]))



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import AffineSystem, Diagram, System


class TestDrakeDiagrams(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(5.)
    def test_actuator_model(self):
        # note: all prints here go to the output item in the json file

        actuator_model = self.notebook_locals['actuator_model']

        assert isinstance(actuator_model,
                          System), "actuator_model must be a System"
        assert actuator_model.ToAutoDiffXdMaybe(
        ), "Your system isn't compatible with Drake's AutoDiffXd; see the note above."  # noqa
        assert actuator_model.num_input_ports(
        ) == 1, "must have exactly one input port"
        assert actuator_model.get_input_port().size(
        ) == 1, "input port must have size 1"
        assert actuator_model.num_output_ports(
        ) == 1, "must have only one output port"
        assert actuator_model.get_output_port().size(
        ) == 1, "output port should have size 1"
        assert actuator_model.num_continuous_states(
        ) == 1, "should have exactly one continuous state variable"
        assert actuator_model.num_discrete_state_groups(
        ) == 0, "should not have any discrete state"

        context = actuator_model.CreateDefaultContext()

        def check_dynamics(x, u):
            context.SetContinuousState([x])
            actuator_model.get_input_port().FixValue(context, [u])
            xdot = actuator_model.EvalTimeDerivatives(context).CopyToVector()
            self.assertAlmostEqual(
                2 * (u - x), xdot[0],
                'the state dynamics (xdot) of your actuator model are incorrect'
            )
            y = actuator_model.get_output_port().Eval(context)
            self.assertAlmostEqual(
                y, x, 'the output value of your actuator model is incorrect')

        check_dynamics(x=0, u=0)
        check_dynamics(x=0.4, u=0.1)
        check_dynamics(x=0.52, u=0.25)
        check_dynamics(x=-0.91, u=5.2)
        check_dynamics(x=2, u=-2)

    @weight(1)
    @timeout_decorator.timeout(5.)
    def test_diagram(self):
        diagram = self.notebook_locals['diagram']
        assert isinstance(diagram, Diagram), 'diagram must be a Diagram'
        assert len(diagram.GetSystems()
                  ) == 2, 'diagram should have exactly two subsystems'  # noqa
        assert diagram.ToAutoDiffXdMaybe(
        ), "Your system isn't compatible with Drake's AutoDiffXd; see the note above."  # noqa
        assert diagram.num_input_ports(
        ) == 1, "diagram should have exactly one input port"
        assert diagram.get_input_port().size(
        ) == 1, "diagram's input port should have size 1"
        assert diagram.num_continuous_states(
        ) == 3, "diagram should have a total of 3 continuous states"
        assert diagram.num_discrete_state_groups(
        ) == 0, "diagram should not have any discrete states"

    @weight(1)
    @timeout_decorator.timeout(5.)
    def test_diagram_context(self):
        diagram = self.notebook_locals['diagram']
        double_integrator = self.notebook_locals['double_integrator']
        actuator_model = self.notebook_locals['actuator_model']
        context = self.notebook_locals['diagram_context']

        double_integrator_context = double_integrator.GetMyContextFromRoot(
            context)
        np.testing.assert_almost_equal(
            double_integrator_context.get_continuous_state_vector().get_value(),
            [1.2, 0],
            err_msg="the value of the double integrator state is incorrect")
        actuator_model_context = actuator_model.GetMyContextFromRoot(context)
        np.testing.assert_almost_equal(
            actuator_model_context.get_continuous_state_vector().get_value(),
            [0],
            err_msg="the value of the actuator model state is incorrect")

        assert diagram.get_input_port().HasValue(
            context), "you must fix a value for the diagram's input port"
        np.testing.assert_almost_equal(
            diagram.get_input_port().Eval(context), [0],
            err_msg="the input port value is incorrect")

    @weight(1)
    @timeout_decorator.timeout(5.)
    def test_controller(self):
        controller = self.notebook_locals['controller']
        assert isinstance(controller,
                          AffineSystem), "controller should be an affine system"

        assert controller.num_continuous_states(
        ) == 0, "your controller has state; that shouldn't happen"
        np.testing.assert_almost_equal(
            controller.D(), [[-1.06519664, -1., -2.26503715]],
            err_msg="your LQR gains do not match the expected value")



import warnings

from underactuated.meshcat_utils import *

warnings.warn(
    "underactuated.meshcat_cpp_utils has been renamed to underactuated.meshcat_utils. This shim will be removed after 2023-05-31.",
    DeprecationWarning,
    stacklevel=2)



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np
from pydrake.all import (DiagramBuilder, SceneGraph, Simulator, LeafSystem,
                         DiagramBuilder, SceneGraph, ZeroOrderHold)
from pydrake.examples import PendulumGeometry, PendulumPlant


class TestPendulumCVI(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(2)
    @timeout_decorator.timeout(20.)
    def test_state_cost(self):
        """Test compute state cost"""
        # note: all prints here go to the output item in the json file

        # the loss must be close to the correct value
        compute_state_cost = self.notebook_locals['compute_state_cost']
        tol = 1e-8

        # intentionally different values than for cartpole
        target_state = np.array([0, np.pi]).reshape(-1, 1)
        Q = np.diag([1., 20.])

        self.assertLessEqual(
            compute_state_cost(Q, target_state, target_state),
            tol,
            msg=('target state cost is '
                 f'{compute_state_cost(Q, target_state, target_state)}'
                 ' which is more than tolerance {tol}'))

        cost = compute_state_cost(Q, target_state,
                                  np.array([0, 0]).reshape(-1, 1))
        target = np.pi**2 * 20.
        self.assertLessEqual(np.abs(cost - target),
                             tol,
                             msg='Cost at state = 0 is incorrect')

        state1 = [
            5.89375203, 5.72132364, 2.07924145, -25.58778621, 4.84281921,
            -19.20542969, -2.35612669
        ]
        state2 = [
            -35.69553553, 24.11618679, -0.85857409, 3.61505439, -3.28412933,
            26.96593774, -0.81457745
        ]
        test_states = np.array([state1, state2])
        result_true = np.array([
            30201.18682396, 8831.40552794, 324.34992454, 659.21812345,
            849.25095811, 11720.8369055, 318.57697075
        ])
        result_tested = compute_state_cost(Q, target_state, test_states)
        for i in range(result_true.shape[0]):
            self.assertLessEqual(np.abs(result_true[i] - result_tested[i]), tol,
                                 f"failed on random state number {i}")

    @weight(2)
    @timeout_decorator.timeout(20.)
    def test_compute_u_star(self):
        """Test compute u_star"""
        # note: all prints here go to the output item in the json file
        compute_u_star = self.notebook_locals['compute_u_star']

        R_diag = np.array([1.5])

        dJdX = np.array(
            [[8.455e+00, -2.330e+00, 1.650e-01, 2.040e+00, -3.945e+00],
             [1.000e-02, -5.000e-03, -8.775e+00, 5.090e+00, 3.000e+00]])
        dstate_dynamics_du = np.array([[[-0.785, 3.86, -6.79, 3.51, 5.265]],
                                       [[1.35, -4.855, 5.24, 6.095, -0.26]]])
        u_star_true = np.array(
            [[2.20789167, 2.98984167, 15.70045, -12.72798333, 7.183475]])
        u_star_test = compute_u_star(R_diag, dJdX, dstate_dynamics_du)
        self.assertTrue(u_star_true.shape == u_star_test.shape,
                        (f"output of compute_u_star is incorrect. Expected "
                         f"{u_star_true.shape} got {u_star_test.shape}"))

        for i in range(u_star_test.shape[0]):
            for j in range(u_star_test.shape[1]):
                self.assertLessEqual(
                    np.abs(u_star_test[i, j] - u_star_true[i, j]), 1e-8,
                    f"incorrect u_star at index {i} for test state {j}")

    @weight(6)
    @timeout_decorator.timeout(600.)
    def test_policy(self):
        """Test policy"""
        value_mlp = self.notebook_locals['value_mlp']
        input_limits = self.notebook_locals['input_limits']
        value_mlp_context = self.notebook_locals['value_mlp_context']
        R_diag = self.notebook_locals['R_diag']
        compute_u_star = self.notebook_locals['compute_u_star']
        simulator = self.build_pendulum_simulator(value_mlp,
                                                  value_mlp_context,
                                                  R_diag,
                                                  compute_u_star,
                                                  input_limits=input_limits)
        simulator_context = simulator.get_mutable_context()
        simulator.set_target_realtime_rate(0.)
        num_sim = 20
        np.random.seed(14)
        errors = []
        for i in range(num_sim):
            duration = 5.0
            simulator_context.SetTime(0.)
            simulator_context.SetContinuousState(
                np.array([2 * np.pi * np.random.rand(), 0]))
            simulator.Initialize()
            simulator.AdvanceTo(duration)
            final_state_vector = simulator_context.get_continuous_state(
            ).get_vector().CopyToVector()
            theta_wrapped = np.arccos(np.cos(final_state_vector[0]))
            errors.append(np.abs(theta_wrapped - np.pi))

        num_passed = 0
        for e in errors:
            # check if theta is close to pi
            if np.abs(e) <= 1e-1:
                num_passed += 1
        frac = 0.7
        threshold = int(frac * num_sim)
        self.assertGreaterEqual(
            num_passed, threshold,
            (f"Only passed {num_passed/num_sim * 100}% of initial"
             f" configurations. Need to pass {frac*100}%"))

    # initialize controller and plant
    def build_pendulum_simulator(self,
                                 value_mlp,
                                 value_mlp_context,
                                 R_diag,
                                 compute_u_star,
                                 input_limits=None):
        time_step = 0.01
        closed_loop_builder = DiagramBuilder()
        plant_cl, scene_graph_cl = closed_loop_builder.AddSystem(
            PendulumPlant()), closed_loop_builder.AddSystem(SceneGraph())

        controller_sys = ContinuousFittedValueIterationPolicyComputeUStar(
            plant_cl,
            value_mlp,
            value_mlp_context,
            R_diag,
            compute_u_star,
            input_limits=input_limits)

        PendulumGeometry.AddToBuilder(closed_loop_builder,
                                      plant_cl.get_state_output_port(),
                                      scene_graph_cl)

        controller = closed_loop_builder.AddSystem(controller_sys)
        # we assume a zero-order hold between time steps
        zoh = closed_loop_builder.AddSystem(ZeroOrderHold(time_step, 1))

        # wire all the systems together
        closed_loop_builder.Connect(plant_cl.get_output_port(),
                                    controller.get_input_port())
        closed_loop_builder.Connect(controller.get_output_port(),
                                    zoh.get_input_port())
        closed_loop_builder.Connect(zoh.get_output_port(),
                                    plant_cl.get_input_port())

        diagram_closed_loop = closed_loop_builder.Build()

        simulator = Simulator(diagram_closed_loop)
        return simulator


class ContinuousFittedValueIterationPolicyComputeUStar(LeafSystem):

    def __init__(self,
                 plant,
                 value_mlp,
                 value_mlp_context,
                 R_diag,
                 compute_u_star,
                 input_port_index=0,
                 input_limits=None):
        LeafSystem.__init__(self)

        self.num_plant_states = value_mlp.get_input_port().size()
        self._plant = plant
        self._plant_context = plant.CreateDefaultContext()

        self.value_mlp = value_mlp
        self.value_mlp_context = value_mlp_context
        self.J = np.zeros((1, 1))
        self.dJdX = np.asfortranarray(np.zeros((self.num_plant_states, 1)))

        self.compute_u_star = compute_u_star

        self.Rinv = 1 / R_diag
        self.R_diag = R_diag
        self.input_limits = input_limits
        self.DeclareVectorInputPort("plant_state", self.num_plant_states)
        self._plant_input_port = self._plant.get_input_port(input_port_index)
        self.DeclareVectorOutputPort("output", self._plant_input_port.size(),
                                     self.CalcOutput)

    def CalcOutput(self, context, output):
        num_inputs = self._plant_input_port.size()
        u = np.zeros(num_inputs)
        plant_state = self.get_input_port().Eval(context)

        self.value_mlp.BatchOutput(self.value_mlp_context,
                                   np.atleast_2d(plant_state).T, self.J,
                                   self.dJdX)

        self._plant_context.SetContinuousState(plant_state)
        self._plant_input_port.FixValue(self._plant_context, u)
        state_dynamics_x = self._plant.EvalTimeDerivatives(
            self._plant_context).CopyToVector()

        dstate_dynamics_du = np.empty((self.num_plant_states, num_inputs, 1))
        for i in range(num_inputs):
            u[i] = 1
            self._plant_input_port.FixValue(self._plant_context, u)
            dstate_dynamics_du[:, :, i] = (self._plant.EvalTimeDerivatives(
                self._plant_context).CopyToVector() - state_dynamics_x).reshape(
                    -1, 1)
            u[i] = 0

        u_star = self.compute_u_star(self.R_diag, self.dJdX,
                                     dstate_dynamics_du)[:, 0]
        if self.input_limits is not None:
            u_star = np.clip(u_star, self.input_limits[0], self.input_limits[1])
        for i in range(num_inputs):
            output.SetAtIndex(i, u_star[i])



import matplotlib.animation as ani
import matplotlib.pyplot as plt
import numpy as np
from IPython.display import HTML
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D
# pydrake imports
from pydrake.all import (DiagramBuilder, DynamicProgrammingOptions,
                         FittedValueIteration, LinearSystem, LogVectorOutput,
                         Simulator, VectorSystem)


def create_animation(J_grid, pi_grid, mesh):
    # initialize figure for animation plus misc settings
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    plt.tight_layout()
    ax[0].axis('off')
    ax[1].axis('off')

    # cost-to-go plot and policy plots
    # overwrites the subplots axes
    ax[0] = fig.add_subplot(121, projection='3d')
    ax[1] = fig.add_subplot(122, projection='3d')
    ax[0].set_title(r'Optimal cost to go   $J^*(\mathbf{x})$')
    ax[1].set_title(r'Optimal policy   $\pi^*(\mathbf{x})$')
    for axi in ax:
        axi.set_xlabel(r'$q$')
        axi.set_ylabel(r'$\dot q$')

    # helper function for the surface plot
    Q, Qdot = np.meshgrid(mesh['q_grid'], mesh['qdot_grid'])

    def plot_surf(ax, Z):
        return ax.plot_surface(Q, Qdot, Z.T, rstride=1, cstride=1, cmap=cm.jet)

    # first frame of the animation
    J_surf = [plot_surf(ax[0], J_grid[:, :, 0])]
    pi_surf = [plot_surf(ax[1], pi_grid[:, :, 0])]

    # video parameters
    frames = 40  # total number of frames
    duration = 10  # seconds
    interval = duration / (frames - 1) * 1000  # milliseconds

    # initialize title to be modified in the callback
    title = fig.text(.5,
                     .95,
                     "",
                     fontsize='x-large',
                     bbox={'facecolor': 'w'},
                     ha='center')

    # callback function for the animation
    def update_surf(frame, J_grid, J_surf, pi_grid, pi_surf):

        # iteration to plot, - 1 for base zero
        iters = J_grid.shape[2]
        i = int(frame * (iters - 1) / (frames - 1))

        # update cost-to-go and policy
        J_surf[0].remove()
        pi_surf[0].remove()
        J_surf[0] = plot_surf(ax[0], J_grid[:, :, i])
        pi_surf[0] = plot_surf(ax[1], pi_grid[:, :, i])

        # update title with current iteration
        # use base 1 as above
        title.set_text(f'Value iteration {i+1}')

    # create animation
    animate = ani.FuncAnimation(fig,
                                update_surf,
                                frames=frames,
                                interval=interval,
                                fargs=(J_grid, J_surf, pi_grid, pi_surf))

    # play video
    plt.close()  # close any open figure
    return animate


# function that plots the trajectory of the
# double integrator in state space (q_sim vs qdot_sim)
def plot_state_trajectory(q_sim, qdot_sim):

    # draw a white dot for the initial conditions
    plt.scatter(q_sim[0],
                qdot_sim[0],
                s=100,
                edgecolor='k',
                c='w',
                zorder=3,
                label=r'$\mathbf{x}(0)$')

    # black line for the trajectory in time
    plt.plot(q_sim, qdot_sim, color='k', linewidth=2, label=r'$\mathbf{x}(t)$')

    # misc
    plt.xlabel(r'$q$')
    plt.ylabel(r'$\dot q$')
    plt.legend()


# helper function for plot_policy that evaluates
# the controller output at a given state
def evaluate_controller(controller, q, qdot):

    # get context and set system output (= system state)
    context = controller.CreateDefaultContext()
    controller.get_input_port(0).FixValue(context, (q, qdot))

    # compute input for the double integrator
    u = controller.get_output_port(0).Eval(context)[0]

    return u


# function that produces a level plot
# the policy generated by the passed controller
# it used the grid defined by q_grid and qdot_grid
def plot_policy(q_grid, qdot_grid, controller):

    # evaluate the policy on the grid
    Pi_grid = np.array(
        [[evaluate_controller(controller, q, qdot)
          for qdot in qdot_grid]
         for q in q_grid])

    # level plot the level function
    # note the transpose to align the policy to the grid
    plt.contourf(*np.meshgrid(q_grid, qdot_grid), Pi_grid.T, cmap=cm.jet)

    # add a bar with the color scale on the right
    plt.colorbar(label=r'$\pi^*(\mathbf{x})$')


# function that plots the control signal
# u_sim as a function of the time vector t
def plot_input(u_sim, t_sim, u_lim):

    # plot the bounds for the control signal
    plt.plot(t_sim, [u_lim[0]] * len(t_sim),
             c='r',
             linestyle='--',
             label='Input limits')
    plt.plot(t_sim, [u_lim[1]] * len(t_sim), c='r', linestyle='--')

    # plot the control signal
    plt.plot(t_sim, u_sim, label='Input from simulation')

    # misc
    plt.xlabel(r'$t$')
    plt.ylabel(r'$u$')
    plt.xlim(min(t_sim), max(t_sim))
    plt.grid(True)
    plt.legend(loc=1)


# overall plot function for the state trajectory,
# controller policy, and input signal
def simulate_and_plot(q0,
                      qdot0,
                      sim_time,
                      controller,
                      u_lim,
                      nq=201,
                      nqdot=201,
                      simulate=None):

    # get trajectories
    q_sim, qdot_sim, u_sim, t_sim = simulate(q0, qdot0, sim_time, controller)

    # state figure
    plt.figure()
    plot_state_trajectory(q_sim, qdot_sim)

    # plot policy only in a square region that
    # contains tightly the trajectory

    # helper function that computes upper and
    # lower bounds, with a small frame, for the
    # passed signal s
    def frame_signal(s, frame=.1):
        ds = (max(s) - min(s)) * frame
        return [min(s) - ds, max(s) + ds]

    # regrid state space for policy plot
    # this grid must be much finer than the
    # one used for value iteration
    q_grid = np.linspace(*frame_signal(q_sim), nq)
    qdot_grid = np.linspace(*frame_signal(qdot_sim), nqdot)

    # plot the policy from the passed controller
    plot_policy(q_grid, qdot_grid, controller)

    # plot input as a function of time
    plt.figure()
    plot_input(u_sim, t_sim, u_lim)



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestDoubleCartPoleURDF(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_q1_get_single_pendulum_urdf_parameters(self):
        """Test the single_pendulum_urdf_parameters"""
        # note: all prints here go to the output item in the json file
        get_single_pendulum_urdf_parameters = self.notebook_locals[
            'get_single_pendulum_urdf_parameters']
        urdf_parameters = get_single_pendulum_urdf_parameters()
        a = 0
        b = 0
        for x in get_single_pendulum_urdf_parameters():
            a += hash(tuple(x))
        for x in [[-1, 0]]:
            b += hash(tuple(x))
        self.assertEqual(a, b, "Incorrect single pendulum urdf parameters.")

    @weight(5)
    @timeout_decorator.timeout(1.)
    def test_q2_get_double_pendulum_urdf_parameters(self):
        """Test the double_pendulum_urdf_parameters"""
        # note: all prints here go to the output item in the json file
        get_double_pendulum_urdf_parameters = self.notebook_locals[
            'get_double_pendulum_urdf_parameters']
        urdf_parameters = get_double_pendulum_urdf_parameters()
        a = 0
        b = 0
        for x in get_double_pendulum_urdf_parameters():
            a += hash(tuple(x))
        for x in [[-1, 0], [1, -1]]:
            b += hash(tuple(x))
        self.assertEqual(a, b, "Incorrect double pendulum urdf parameters.")



from .jupyter import running_as_notebook
from .pyplot_visualizer import SliderSystem
from .multibody import ManipulatorDynamics
# Note: can't import running_as_test here.  Will get a copy instead of a
# reference.
from .utils import FindResource
from .plot_utils import plot_2d_phase_portrait



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestOrbitalTransfer(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    def trajopt_feasibility(self):
        """Test feasibility of the trajopt solution"""
        # load locals
        universe = self.notebook_locals['universe']
        thrust = self.notebook_locals['thrust_opt']
        state = self.notebook_locals['state_opt']
        time_steps = self.notebook_locals['time_steps']
        time_interval = self.notebook_locals['time_interval']

        # initial state in Earth orbit
        residuals = universe.constraint_state_to_orbit(state[0], 'Earth')
        np.testing.assert_array_almost_equal(
            residuals,
            np.zeros(residuals.shape),
            err_msg='The initial state in not on the Earth orbit.')

        # final state in Mars orbit
        residuals = universe.constraint_state_to_orbit(state[-1], 'Mars')
        np.testing.assert_array_almost_equal(
            residuals,
            np.zeros(residuals.shape),
            # decimal=4,
            err_msg='The final state in not on the Mars orbit.')

        # discretized dynamics
        for t in range(time_steps):
            residuals = universe.rocket_discrete_dynamics(
                state[t], state[t + 1], thrust[t], time_interval)
            np.testing.assert_array_almost_equal(
                residuals,
                np.zeros(residuals.shape),
                err_msg=f'Discrete dynamics at time step {t} is not verified.')

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_thrust_limits(self):
        """Test thrust limits"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals['universe']
        thrust = self.notebook_locals['thrust_opt']
        time_steps = self.notebook_locals['time_steps']

        # check thrust limits
        tol = 1e-3
        thrust_limit = universe.rocket.thrust_limit
        for t in range(time_steps):
            self.assertTrue(np.linalg.norm(thrust[t]) <= thrust_limit + tol,
                            msg=f'Thrust limit violated at time step {t}.')

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_velocity_limits(self):
        """Test velocity limits"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals['universe']
        velocity = self.notebook_locals['state_opt'][:, 2:]
        time_steps = self.notebook_locals['time_steps']

        # check velocity limits
        tol = 1e-3
        velocity_limit = universe.rocket.velocity_limit
        for t in range(time_steps + 1):
            self.assertTrue(np.linalg.norm(velocity[t]) <= velocity_limit + tol,
                            msg=f'Velocity limit violated at time step {t}.')

    @weight(6)
    @timeout_decorator.timeout(1.)
    def test_asteroid_collision_limits(self):
        """Test asteroid collision"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals['universe']
        state = self.notebook_locals['state_opt']
        time_steps = self.notebook_locals['time_steps']
        n_asteroids = self.notebook_locals['n_asteroids']

        # avoid collision with asteroids
        tol = 1e-3
        for i in range(n_asteroids):
            asteroid_orbit = universe.get_planet(f'Asteroid_{i}').orbit
            for t in range(time_steps + 1):
                p = universe.position_wrt_planet(state[t], f'Asteroid_{i}')
                self.assertTrue(
                    p.dot(p) >= asteroid_orbit**2 - tol,
                    msg=f'Collision with asteroid {i} at time step {t}.')

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_fuel_consumption_limits(self):
        """Test fuel consumption"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals['universe']
        thrust = self.notebook_locals['thrust_opt']
        time_interval = self.notebook_locals['time_interval']

        # check maximum consumption
        consumption = time_interval * sum(t.dot(t) for t in thrust)
        self.assertTrue(
            consumption <= 250,
            msg=f'Fuel consumption is {consumption}, greater than 250.')



import numpy as np

from pydrake.systems.framework import LeafSystem_, PortDataType
from pydrake.systems.pyplot_visualizer import PyPlotVisualizer
from pydrake.systems.scalar_conversion import TemplateSystem

# Note: In order to use the Python system with drake's autodiff features, we
# have to add a little "TemplateSystem" boilerplate (for now).  For details,
# see https://drake.mit.edu/pydrake/pydrake.systems.scalar_conversion.html


# TODO(russt): Clean this up pending any resolutions on
#  https://github.com/RobotLocomotion/drake/issues/10745
@TemplateSystem.define("Quadrotor2D_")
def Quadrotor2D_(T):

    class Impl(LeafSystem_[T]):

        def _construct(self, converter=None):
            LeafSystem_[T].__init__(self, converter)
            # two inputs (thrust)
            self.DeclareVectorInputPort("u", 2)
            # three positions, three velocities
            state_index = self.DeclareContinuousState(3, 3, 0)
            # six outputs (full state)
            self.DeclareStateOutputPort("x", state_index)

            # parameters based on [Bouadi, Bouchoucha, Tadjine, 2007]
            self.length = 0.25  # length of rotor arm
            self.mass = 0.486  # mass of quadrotor
            self.inertia = 0.00383  # moment of inertia
            self.gravity = 9.81  # gravity

        def _construct_copy(self, other, converter=None):
            Impl._construct(self, converter=converter)

        def DoCalcTimeDerivatives(self, context, derivatives):
            x = context.get_continuous_state_vector().CopyToVector()
            u = self.EvalVectorInput(context, 0).CopyToVector()
            q = x[:3]
            qdot = x[3:]
            qddot = np.array([
                -np.sin(q[2]) / self.mass * (u[0] + u[1]),
                np.cos(x[2]) / self.mass * (u[0] + u[1]) - self.gravity,
                self.length / self.inertia * (u[0] - u[1])
            ])
            derivatives.get_mutable_vector().SetFromVector(
                np.concatenate((qdot, qddot)))

    return Impl


Quadrotor2D = Quadrotor2D_[None]  # Default instantiation


class Quadrotor2DVisualizer(PyPlotVisualizer):

    def __init__(self, ax=None, show=None):
        PyPlotVisualizer.__init__(self, ax=ax, show=show)
        self.DeclareInputPort("state", PortDataType.kVectorValued, 6)
        self.ax.set_aspect("equal")
        self.ax.set_xlim(-2, 2)
        self.ax.set_ylim(-1, 1)

        self.length = .25  # moment arm (meters)

        self.base = np.vstack((1.2 * self.length * np.array([1, -1, -1, 1, 1]),
                               0.025 * np.array([1, 1, -1, -1, 1])))
        self.pin = np.vstack((0.005 * np.array([1, 1, -1, -1, 1]),
                              .1 * np.array([1, 0, 0, 1, 1])))
        a = np.linspace(0, 2 * np.pi, 50)
        self.prop = np.vstack(
            (self.length / 1.5 * np.cos(a), .1 + .02 * np.sin(2 * a)))

        # yapf: disable
        self.base_fill = self.ax.fill(
            self.base[0, :], self.base[1, :], zorder=1, edgecolor="k",
            facecolor=[.6, .6, .6])
        self.left_pin_fill = self.ax.fill(
            self.pin[0, :], self.pin[1, :], zorder=0, edgecolor="k",
            facecolor=[0, 0, 0])
        self.right_pin_fill = self.ax.fill(
            self.pin[0, :], self.pin[1, :], zorder=0, edgecolor="k",
            facecolor=[0, 0, 0])
        self.left_prop_fill = self.ax.fill(
            self.prop[0, :], self.prop[0, :], zorder=0, edgecolor="k",
            facecolor=[0, 0, 1])
        self.right_prop_fill = self.ax.fill(
            self.prop[0, :], self.prop[0, :], zorder=0, edgecolor="k",
            facecolor=[0, 0, 1])
        # yapf: enable

    def draw(self, context):
        x = self.EvalVectorInput(context, 0).CopyToVector()
        R = np.array([[np.cos(x[2]), -np.sin(x[2])],
                      [np.sin(x[2]), np.cos(x[2])]])

        p = np.dot(R, self.base)
        self.base_fill[0].get_path().vertices[:, 0] = x[0] + p[0, :]
        self.base_fill[0].get_path().vertices[:, 1] = x[1] + p[1, :]

        p = np.dot(R, np.vstack(
            (-self.length + self.pin[0, :], self.pin[1, :])))
        self.left_pin_fill[0].get_path().vertices[:, 0] = x[0] + p[0, :]
        self.left_pin_fill[0].get_path().vertices[:, 1] = x[1] + p[1, :]
        p = np.dot(R, np.vstack((self.length + self.pin[0, :], self.pin[1, :])))
        self.right_pin_fill[0].get_path().vertices[:, 0] = x[0] + p[0, :]
        self.right_pin_fill[0].get_path().vertices[:, 1] = x[1] + p[1, :]

        p = np.dot(R,
                   np.vstack((-self.length + self.prop[0, :], self.prop[1, :])))
        self.left_prop_fill[0].get_path().vertices[:, 0] = x[0] + p[0, :]
        self.left_prop_fill[0].get_path().vertices[:, 1] = x[1] + p[1, :]

        p = np.dot(R, np.vstack(
            (self.length + self.prop[0, :], self.prop[1, :])))
        self.right_prop_fill[0].get_path().vertices[:, 0] = x[0] + p[0, :]
        self.right_prop_fill[0].get_path().vertices[:, 1] = x[1] + p[1, :]

        self.ax.set_title("t = {:.1f}".format(context.get_time()))



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np
from pydrake.all import Solve


class TestVanDerPol(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(5.)
    def test_method_1(self):
        """Test level set for Method 1"""
        # test level set value up to first digit
        rho_method_1 = self.notebook_locals['rho_method_1']
        self.assertEqual(hash(self.round(rho_method_1, 1)), 691752902764107778,
                         "The level set from Method 1 is incorrect.")

        # test that the line search converged
        is_verified = self.notebook_locals['is_verified']
        self.assertTrue(
            is_verified(rho_method_1),
            "Could not verify the level set rho_method_1 with the function"
            + "is_verified(rho).")
        rho_step = self.notebook_locals['rho_step']
        self.assertAlmostEqual(
            rho_step,
            0.01,
            msg="rho_step is different from the predefined value of 0.01.")
        self.assertFalse(
            is_verified(rho_method_1 + rho_step),
            "The function is_verified(rho) could verify the level set"
            + "rho_method_1 + rho_step: line search is not complete.")

    @weight(5)
    @timeout_decorator.timeout(5.)
    def test_method_2(self):
        """Test level set for Method 2"""
        # test level set value up to fourth digit
        # (note that the degree of the polynomial does not affect precision
        # in this particular problem)
        rho_method_2 = self.notebook_locals['rho_method_2']
        self.assertEqual(hash(self.round(rho_method_2, 4)), 702129196305569794,
                         "The level set from Method 2 is incorrect.")

        # try to solve again the optimization problem
        # and test that the optimal value of rho is the same
        # to retrieve rho, find the only optimization variable in the cost
        prog = self.notebook_locals['prog2']
        costs = prog.GetAllCosts()
        self.assertEqual(len(costs), 1,
                         "prog2 has more than one cost function.")
        variables = prog.GetAllCosts()[0].variables()
        self.assertEqual(len(variables), 1,
                         "The cost function of prog2 is incorrect.")
        rho = variables[0]
        result = Solve(prog)
        self.assertAlmostEqual(
            result.GetSolution(rho),
            rho_method_2,
            msg="Solving prog2 we got a different value of rho_method_2.")

    @weight(5)
    @timeout_decorator.timeout(5.)
    def test_method_3(self):
        """Test level set for Method 3"""
        # test level set value up to fourth digit
        # (note that the degree of the polynomial does not affect precision
        # in this particular problem)
        rho_method_3 = self.notebook_locals['rho_method_3']
        self.assertEqual(hash(self.round(rho_method_3, 4) * 2),
                         1404258392611139588,
                         "The level set from Method 3 is incorrect.")

        # try to solve again the optimization problem
        # and test that the optimal value of rho is the same
        # to retrieve rho, find the only optimization variable in the cost
        prog = self.notebook_locals['prog3']
        costs = prog.GetAllCosts()
        self.assertEqual(len(costs), 1,
                         "prog3 has more than one cost function.")
        variables = prog.GetAllCosts()[0].variables()
        self.assertEqual(len(variables), 1,
                         "The cost function of prog3 is incorrect.")
        rho = variables[0]
        result = Solve(prog)
        self.assertAlmostEqual(
            result.GetSolution(rho),
            rho_method_3,
            msg="Solving prog3 we got a different value of rho_method_3.")

    @staticmethod
    def round(n, decimals=0):
        multiplier = 10**decimals
        return round(n * multiplier) / multiplier



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestOneDHopper(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_mechanical_energy(self):
        """Test mechanical energy"""
        # retrieve energy function
        mechanical_energy = self.notebook_locals['mechanical_energy']

        # evaluate energy at N random points
        N = 100
        np.random.seed(0)
        state = np.random.randn(4, N)
        delta = np.random.randn(N)
        E = mechanical_energy(state, delta)

        # check size of E
        self.assertFalse(
            isinstance(E, (float, int)),
            msg=f'The function mechanical_energy does not take vectorized '
            + 'inputs.')
        self.assertTrue(
            E.size == N,
            msg=f'Passing an array of shape (4, {N}) for the states and an '
            + f'array of shape ({N},) for delta, got back an array of size '
            + f'{E.size} from the function mechanical_energy. The output array '
            + f'must have size {N} instead.')

        # compare with target values
        E_target = np.array([
            1035.88403634, 3066.05294704, 226.59829248, 1132.19808917,
            4519.60941109, 2571.77722312, 179.75301624, 218.267293,
            1518.71951559, 285.28744414, 126.93879604, 1623.09759602,
            761.6724944, 1439.16285815, 1123.31075613, 1150.29073766,
            1039.18526308, -22.85241161, 2344.88095522, 3406.4444863,
            586.89991222, 194.75259958, 1710.76292998, 985.57408993,
            593.67007518, -156.80729216, 129.76883127, 397.72717689,
            179.93169091, 392.49996925, 784.32597859, 433.87021831,
            -81.97555113, -164.62661332, 1261.32659852, 518.81015917,
            1127.06741275, 171.04599127, 678.92178616, 476.58699247,
            1653.76175357, 111.44797797, -178.80377977, 562.21356062,
            5279.60802163, 321.88033307, -33.49681192, 113.06257362,
            2018.19847436, 147.46771374, 534.50643625, 3271.78401092,
            -21.5353477, 15.30134453, 27.29342195, 963.63275164, 240.49348165,
            5417.05008921, 384.4609627, 78.46055822, 445.4086912, 433.45750051,
            53.56942879, -149.21833514, 1521.20337472, 5268.5937194,
            1737.91194436, 2767.73189444, 520.04810383, 170.84779081,
            688.84932501, 86.93212697, 323.27284373, 786.95926209, 283.83230408,
            1338.02751306, 33.00346265, 1167.80440843, 211.53932223,
            163.70938173, -40.39089787, 149.60356967, 1255.61900843,
            3329.55930887, 1274.00886131, 9255.17733426, 207.51077886,
            171.65671047, 190.63432638, 1693.4769287, 2643.46635344,
            139.37393331, 613.39405767, 111.57079258, 5226.36517526,
            111.22451849, 15.62008105, 251.80124408, 692.01834266, 850.05120716
        ])
        np.testing.assert_array_almost_equal(
            E_target,
            E.flatten(),
            err_msg='The function mechanical_energy does not return the '
            + 'correct values when evaluated at random points.')

    @weight(10)
    @timeout_decorator.timeout(15.)
    def test_preload_controller(self):
        """Test preload controller"""
        # retrieve controller and simulator
        PreloadController = self.notebook_locals['PreloadController']
        simulate = self.notebook_locals['simulate']

        # number of random simulations
        N = 10

        # duration of each simulation, just enough to ensure
        # convergence independently of the initial condition
        duration = 6

        # final window of time in which convergence must be achieved
        tail = 2

        np.random.seed(1)
        for i in range(N):

            # desired hopping height from .5 to 1.5
            h = np.random.rand() + .5

            # initial body height from 0 to 1
            zb0 = np.random.rand()
            initial_state = np.array([zb0, 0, 0, 0])

            # instantiate controller
            preload_controller = PreloadController(h)

            # run simulation
            logger = simulate(initial_state, duration, preload_controller)[1]

            # unpack logger
            time = logger.sample_times()
            zb = logger.data()[0]

            # get body height for the final window of time
            zb_tail = zb[np.where(time >= duration - tail)[0]]

            # ensure that in final 2 seconds of simulation the body
            # is always below h plus a tolerance
            tol = .1
            self.assertTrue(
                max(zb_tail) < h + tol,
                msg=f'Starting from the state {initial_state} and having a '
                + f'target hopping height of {h}, the hopper body still hops '
                + f'above {h} after {duration - tail} seconds.')

            # ensure that in final 2 seconds of simulation the body achieves
            # the hopping height h at least once
            self.assertTrue(
                max(zb_tail) > h - tol,
                msg=f'Starting from the state {initial_state} and having a '
                + f'target hopping height of {h}, the hopper body does not '
                + f'reach the height {h} after {duration - tail} seconds.')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestGliderSysid(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(20.)
    def test_lq_fit(self):
        """Test least squares fitting"""
        # note: all prints here go to the output item in the json file

        # the best config and best score should have the right values
        best_config = self.notebook_locals['best_config']
        best_score = self.notebook_locals['best_score']
        self.assertFalse(best_config != 6, 'The chosen basis set is incorrect')
        self.assertFalse(best_score != 63831, 'The best score is incorrect')

    @weight(3)
    @timeout_decorator.timeout(20.)
    def test_basis_dynamics(self):
        """Test basis function dynamics"""
        # note: all prints here go to the output item in the json file
        trajectories_fit = self.notebook_locals['trajectories_fit']
        last_x = trajectories_fit[0][-1][0]
        last_z = trajectories_fit[0][-1][1]
        last_th = trajectories_fit[0][-1][2]
        self.assertLessEqual(last_x,
                             3.36,
                             msg='Simulated trajectory x is incorrect')
        self.assertGreaterEqual(last_x,
                                3.34,
                                msg='Simulated trajectory x is incorrect')
        self.assertLessEqual(last_z,
                             5.10,
                             msg='Simulated trajectory z is incorrect')
        self.assertGreaterEqual(last_z,
                                4.99,
                                msg='Simulated trajectory z is incorrect')
        self.assertLessEqual(last_th,
                             -0.52,
                             msg='Simulated trajectory theta is incorrect')
        self.assertGreaterEqual(last_th,
                                -0.54,
                                msg='Simulated trajectory theta is incorrect')



import numpy as np
import math
import matplotlib.animation as animation
from pydrake.systems.framework import Context, PortDataType
from pydrake.systems.pyplot_visualizer import PyPlotVisualizer


class PendulumVisualizer(PyPlotVisualizer):
    a1 = 0.75
    ac1 = 0.75
    av = np.linspace(0, math.pi, 20)
    rb = .03
    hb = .07
    aw = .01
    base_x = rb * np.concatenate(([1], np.cos(av), [-1]))
    base_y = np.concatenate(([-hb], rb * np.sin(av), [-hb]))
    arm_x = np.concatenate(
        (aw * np.sin(av - math.pi / 2), aw * np.sin(av + math.pi / 2)))
    arm_y = np.concatenate(
        (aw * np.cos(av - math.pi / 2), -a1 + aw * np.cos(av + math.pi / 2)))

    def __init__(self, ax=None, show=None):
        PyPlotVisualizer.__init__(self, ax=ax, show=show)
        self.set_name("pendulum_visualizer")
        self.DeclareInputPort("state", PortDataType.kVectorValued, 2)

        self.ax.set_xlim([-1.2, 1.2])
        self.ax.set_ylim([-1.2, 1.2])
        # yapf: disable
        self.base = self.ax.fill(
            self.base_x, self.base_y, zorder=1, facecolor=(.3, .6, .4),
            edgecolor="k")

        # arm_x and arm_y are closed (last element == first element), but don't
        # pass the last element to the fill command, because it gets closed
        # anyhow (and we want the sizes to match for the update).
        self.arm = self.ax.fill(
            self.arm_x[0:-1], self.arm_y[0:-1], zorder=0, facecolor=(.9, .1, 0),
            edgecolor="k")
        self.center_of_mass = self.ax.plot(
            0, -self.ac1, zorder=1, color="b", marker="o", markersize=14)
        # yapf: enable

    def draw(self, context):
        if isinstance(context, Context):
            theta = self.EvalVectorInput(context, 0).get_value()[0]
            self.ax.set_title("t = {:.1f}".format(context.get_time()))
        else:
            theta = context[0]
            self.ax.set_title("")

        path = self.arm[0].get_path()
        path.vertices[:, 0] = self.arm_x * math.cos(
            theta) - self.arm_y * math.sin(theta)
        path.vertices[:, 1] = self.arm_x * math.sin(
            theta) + self.arm_y * math.cos(theta)
        self.center_of_mass[0].set_data(self.ac1 * math.sin(theta),
                                        -self.ac1 * math.cos(theta))
